/***************************************************************************
 * Copyright 2012 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.business.beans.request;

import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

/**
 * This simple bean can be used to store a string during a request. Furthermore it provides some simple methods to handle strings within the application where a bean
 * is necessary. <br>
 * For technical reasons this bean is a request scoped bean.
 * 
 * @author Nils Christian Ehmke
 */
@Component
@Scope("request")
public final class StringBean {

	private String string;

	/**
	 * Creates a new instance of this bean and initializes it with an empty string. <b>Do not use this constructor. This bean is Spring managed.</b>
	 */
	public StringBean() {
		this.string = "";
	}

	/**
	 * Delivers the value of this bean.
	 * 
	 * @return The current string stored within this bean.
	 */
	public String getString() {
		return this.string;
	}

	/**
	 * Sets the value of this bean.
	 * 
	 * @param string
	 *            The new string to be stored within this bean.
	 */
	public void setString(final String string) {
		this.string = string;
	}

	/**
	 * This method verifies whether the given object is an instance of {@link java.lang.String} or not. This is necessary due to the fact that the
	 * <i>instanceof</i>-command has not yet been implemented in JSF.
	 * 
	 * @param object
	 *            The object to be verified.
	 * @return true if and only if the given object is an instance of String.
	 */
	public boolean checkString(final Object object) {
		return object instanceof String;
	}

	/**
	 * Shortens a given name. If the name is longer than the given number of characters, only the first maxChar - 3 characters of the string will be used and the
	 * last three will be replaced with dots.
	 * 
	 * @param name
	 *            The name to be shortened.
	 * @param maxChar
	 *            The maximal number of characters.
	 * @return The shortened name.
	 */
	public String shortenLongName(final String name, final int maxChar) {
		if (name.length() > maxChar) {
			return name.substring(0, maxChar - 3) + "...";
		} else {
			return name;
		}
	}
}
