/***************************************************************************
 * Copyright 2012 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.business.beans.session;

import java.io.Serializable;
import java.util.Map;

import javax.faces.context.FacesContext;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletResponse;

import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Scope;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.stereotype.Component;

import kieker.webgui.business.beans.application.GlobalPropertiesBean;

/**
 * This bean contains information about the user of this session (like the properties and configurations). This class is a Spring managed bean with session scope.
 * This means also that it is possible to login the same user multiple times. As it is a session bean, it must implement the {@link Serializable} interface.
 * 
 * @author Nils Christian Ehmke
 */
@Component
@Scope("session")
public final class UserBean implements Serializable, InitializingBean {

	private static final long serialVersionUID = 6422563876003638348L;

	private String lookAndFeel;
	private String gridColor;
	private int gridSize;

	@Autowired
	private GlobalPropertiesBean globalPropertiesBean;

	/**
	 * Default constructor. <b>Do not use this constructor. This bean is Spring managed.</b>
	 */
	public UserBean() {
		// No code necessary
	}

	/**
	 * Returns the name of the user of the current session. If something goes wrong during the search, it returns a human readable 'N/A'.
	 * 
	 * @return The user name of the session user.
	 */
	public String getUsername() {
		final String username;

		// Try to get the principal
		final Object principal = SecurityContextHolder.getContext().getAuthentication().getPrincipal();
		if (principal instanceof UserDetails) {
			// Looks like it worked. Extract the username.
			final UserDetails userDetails = (UserDetails) principal;
			username = userDetails.getUsername();
		} else {
			// Something went wrong
			username = "N/A";
		}

		return username;
	}

	/**
	 * The getter for the property {@link UserBean#lookAndFeel}.
	 * 
	 * @return The current value for the property.
	 */
	public String getLookAndFeel() {
		synchronized (this) {
			return this.lookAndFeel;
		}
	}

	/**
	 * The setter for the property {@link UserBean#lookAndFeel}.
	 * 
	 * @param lookAndFeel
	 *            The new value for the property.
	 */
	public void setLookAndFeel(final String lookAndFeel) {
		synchronized (this) {
			this.lookAndFeel = lookAndFeel;
			UserBean.saveValueInCookie(this.globalPropertiesBean.getThemeCookieName(), lookAndFeel);
		}
	}

	/**
	 * The getter for the property {@link UserBean#gridColor}.
	 * 
	 * @return The current value for the property.
	 */
	public String getGridColor() {
		synchronized (this) {
			return this.gridColor;
		}
	}

	/**
	 * The setter for the property {@link UserBean#gridColor}.
	 * 
	 * @param gridColor
	 *            The new value for the property.
	 */
	public void setGridColor(final String gridColor) {
		synchronized (this) {
			this.gridColor = gridColor;
			UserBean.saveValueInCookie(this.globalPropertiesBean.getAnalysisEditorGridColorCookieName(), gridColor);
		}
	}

	/**
	 * The getter for the property {@link UserBean#gridSize}.
	 * 
	 * @return The current value for the property.
	 */
	public int getGridSize() {
		synchronized (this) {
			return this.gridSize;
		}
	}

	/**
	 * The setter for the property {@link UserBean#gridSize}.
	 * 
	 * @param gridSize
	 *            The new value for the property.
	 */
	public void setGridSize(final int gridSize) {
		synchronized (this) {
			this.gridSize = gridSize;
			UserBean.saveValueInCookie(this.globalPropertiesBean.getAnalysisEditorGridSizeCookieName(), Integer.toString(gridSize));
		}
	}

	/**
	 * This method tries to load the default values of the properties from the {@link GlobalPropertiesBean} and the faces context.
	 */
	private void loadDefaultValues() {
		synchronized (this) {
			// Get the parameters within the current context.
			final Map<String, String> params = FacesContext.getCurrentInstance().getExternalContext().getRequestParameterMap(); // NOPMD (No concurrent access)
			// Try to find the default theme within the parameters.
			if (params.containsKey(this.globalPropertiesBean.getFacesContextThemeKey())) {
				this.lookAndFeel = params.get(this.globalPropertiesBean.getFacesContextThemeKey());
			} else {
				// Use the default theme.
				this.lookAndFeel = this.globalPropertiesBean.getDefaultTheme();
			}

			this.gridColor = this.globalPropertiesBean.getAnalysisEditorDefaultGridColor();
			this.gridSize = Integer.parseInt(this.globalPropertiesBean.getAnalysisEditorDefaultGridSize());
		}
	}

	/**
	 * This method tries to load the values of the properties from the user's cookies.
	 */
	private void loadValuesFromCookies() {
		synchronized (this) {
			// Get the available cookies
			final Map<String, Object> cookies = FacesContext.getCurrentInstance().getExternalContext().getRequestCookieMap(); // NOPMD (No concurrent access)

			// Now load the values if available
			if (cookies.containsKey(this.globalPropertiesBean.getThemeCookieName())) {
				this.lookAndFeel = ((Cookie) cookies.get(this.globalPropertiesBean.getThemeCookieName())).getValue();
			}
			if (cookies.containsKey(this.globalPropertiesBean.getAnalysisEditorGridColorCookieName())) {
				this.gridColor = ((Cookie) cookies.get(this.globalPropertiesBean.getAnalysisEditorGridColorCookieName())).getValue();
			}
			if (cookies.containsKey(this.globalPropertiesBean.getAnalysisEditorGridSizeCookieName())) {
				this.gridSize = Integer.parseInt(((Cookie) cookies.get(this.globalPropertiesBean.getAnalysisEditorGridSizeCookieName())).getValue());
			}
		}
	}

	/**
	 * This method saves the given values in a cookie with the given name, using the maximal available age for the cookie.
	 * 
	 * @param cookieName
	 *            The name of the cookie.
	 * @param cookieValue
	 *            The value for the cookie.
	 */
	private static void saveValueInCookie(final String cookieName, final String cookieValue) {
		// Create the cookie and make sure that it will be saved for a year (maximum age)
		final Cookie cookie = new Cookie(cookieName, cookieValue);
		cookie.setMaxAge(60 * 60 * 24 * 365);

		// Deliver the cookie
		final HttpServletResponse response = (HttpServletResponse) FacesContext.getCurrentInstance().getExternalContext().getResponse();
		response.addCookie(cookie);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.springframework.beans.factory.InitializingBean#afterPropertiesSet()
	 */
	@Override
	public void afterPropertiesSet() throws Exception {
		this.loadDefaultValues();
		this.loadValuesFromCookies();
	}
}
