/***************************************************************************
 * Copyright 2012 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.business.beans.view;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import kieker.webgui.domain.User;
import kieker.webgui.domain.User.Role;
import kieker.webgui.service.IUserService;

/**
 * @author Nils Christian Ehmke
 */
@Component
@Scope("view")
public final class CurrentUserManagementBean implements InitializingBean {

	@Autowired
	private IUserService userService;
	/**
	 * This list contains all available users.
	 */
	private List<User> users = new ArrayList<User>();

	/**
	 * Default constructor. <b>Do not call this constructor manually. It will only be accessed by Spring.</b>
	 */
	public CurrentUserManagementBean() {
		// No code necessary
	}

	/**
	 * This method adds the given user to the database and informs about success via the growl component.
	 * 
	 * @param username
	 *            The name of the new user.
	 * @param password
	 *            The password of the new user.
	 * @param roles
	 *            The roles of the new user.
	 */
	public void addUser(final String username, final String password, final List<Role> roles) {
		this.userService.addUser(new User(username, password, roles, true));
		// TODO Check that the op was successful before adding the user to our list
		this.users.add(new User(username, "", roles, true));
	}

	/**
	 * This method removed the given user from the database and informs about success via the growl component.
	 * 
	 * @param username
	 *            The name of the user.
	 */
	public void deleteUser(final User user) {
		this.userService.deleteUser(user);
	}

	/**
	 * This method edits the given user within the database and informs about success via the growl component.
	 * 
	 * @param username
	 *            The name of the user.
	 * @param password
	 *            The (new) password of the user.
	 * @param roles
	 *            The (new) roles of the user.
	 */
	public void editUser(final String username, final String password, final Role... roles) {
		this.userService.editUserWithPassword(new User(username, password, Arrays.asList(roles), true));
	}

	/**
	 * Updates the list containing the available users.
	 */
	public void updateList() {
		this.users = this.userService.getUsers();
	}

	/**
	 * This method delivers all available users.
	 * 
	 * @return All currently available users.
	 */
	public List<User> getUsers() {
		return this.users;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.springframework.beans.factory.InitializingBean#afterPropertiesSet()
	 */
	@Override
	public void afterPropertiesSet() throws Exception {
		this.updateList();
	}
}
