/***************************************************************************
 * Copyright 2012 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/
package kieker.webgui.persistence.impl;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import javax.sql.DataSource;

import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Scope;
import org.springframework.jdbc.datasource.DataSourceUtils;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.webgui.common.exception.DataAccessException;
import kieker.webgui.domain.User;
import kieker.webgui.domain.User.Role;
import kieker.webgui.persistence.IUserDAO;

/**
 * This is an implementation of the {@link IUserDAO} interface, which uses Apache Derby to store the users. It is a Spring managed singleton bean.
 * 
 * @author Nils Christian Ehmke
 */
// FIXME Transactional annotation don't work yet
@Component
@Scope("singleton")
public class DerbyUserDAO implements IUserDAO, InitializingBean, DisposableBean {

	private static final Log LOG = LogFactory.getLog(DerbyUserDAO.class);

	@Autowired
	private DataSource userDataSource;
	private Connection connection;

	/**
	 * Default constructor. <b>Do not use this constructor. This bean is Spring managed.</b>
	 */
	public DerbyUserDAO() {
		// No code necessary
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.springframework.beans.factory.DisposableBean#destroy()
	 */
	@Override
	public void destroy() throws Exception {
		this.connection.close();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.springframework.beans.factory.InitializingBean#afterPropertiesSet()
	 */
	@Override
	public void afterPropertiesSet() throws Exception {
		this.connection = DataSourceUtils.getConnection(this.userDataSource);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IUserDAO#addUser(kieker.webgui.domain.User)
	 */
	@Override
	@PreAuthorize("hasRole('ROLE_ADMIN')")
	@Transactional
	public void addUser(final User user) throws DataAccessException {
		PreparedStatement userCmd = null;
		PreparedStatement roleCmd = null;

		try {
			// Prepare both statements
			userCmd = this.connection.prepareStatement("INSERT INTO KIEKERUser (name, password, enabled) VALUES (?, ?, ?)");
			roleCmd = this.connection.prepareStatement("INSERT INTO Userroles (name, role) VALUES (?, ?)");

			// Add the user itself to the database
			userCmd.setString(1, user.getName());
			userCmd.setString(2, user.getPassword());
			userCmd.setBoolean(3, user.isEnabled());
			userCmd.execute();

			// Now the roles
			roleCmd.setString(1, user.getName());
			for (final Role role : user.getRoles()) {
				roleCmd.setInt(2, role.getID());
				roleCmd.execute();
			}
		} catch (final SQLException ex) {
			throw new DataAccessException("Could not add user to the database.", ex);
		} finally {
			if (userCmd != null) {
				try {
					userCmd.close();
				} catch (final SQLException ex) {
					// Log the exception, but don't inform the upper layer about this
					DerbyUserDAO.LOG.warn("Could not close prepared statement.", ex);
				}
			}
			if (roleCmd != null) {
				try {
					roleCmd.close();
				} catch (final SQLException ex) {
					// Log the exception, but don't inform the upper layer about this
					DerbyUserDAO.LOG.warn("Could not close prepared statement.", ex);
				}
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IUserDAO#editUser(kieker.webgui.domain.User)
	 */
	@Override
	@PreAuthorize("hasRole('ROLE_ADMIN')")
	@Transactional
	public void editUserWithPassword(final User user) throws DataAccessException {
		// TODO Auto-generated method stub
	}

	@Override
	@PreAuthorize("hasRole('ROLE_ADMIN')")
	@Transactional
	public void editUserWithoutPassword(final User user) throws DataAccessException {
		// TODO Auto-generated method stub
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IUserDAO#deleteUser(kieker.webgui.domain.User)
	 */
	@Override
	@PreAuthorize("hasRole('ROLE_ADMIN')")
	@Transactional
	public void deleteUser(final User user) throws DataAccessException {
		// TODO Auto-generated method stub
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.persistence.IUserDAO#getUsers()
	 */
	@Override
	@PreAuthorize("hasRole('ROLE_ADMIN')")
	@Transactional(readOnly = true)
	public List<User> getUsers() {
		// FIXME Users without roles
		final List<User> result = new ArrayList<User>();
		ResultSet queryResult = null;

		try {
			final Map<String, User> tempMap = new TreeMap<String, User>();

			final PreparedStatement getQuery = this.connection
					.prepareStatement("select u.name, ur.role, u.enabled from KIEKERUser u, Userroles ur where u.name=ur.name");

			// Run through all results
			queryResult = getQuery.executeQuery();
			while (queryResult.next()) {
				// Get both the username and the role from the current entry
				final String username = queryResult.getString(1);
				final int roleID = queryResult.getInt(2);
				final Role role = Role.fromID(roleID);
				final boolean enabled = queryResult.getBoolean(3);

				// If the user doesn't exist in our map yet, add him.
				// In each case we add the role to the user
				if (tempMap.containsKey(username)) {
					tempMap.get(username).getRoles().add(role);
				} else {
					final List<Role> roles = new ArrayList<Role>();
					roles.add(role);
					tempMap.put(username, new User(username, "", roles, enabled));
				}
			}

			// Now convert the map to the list
			result.addAll(tempMap.values());
		} catch (final SQLException ex) {
			throw new DataAccessException("Could not receive user list.", ex);
		} finally {
			try {
				if (queryResult != null) {
					queryResult.close();
				}
			} catch (final SQLException ex) {
				// Log the exception, but don't inform the upper layer about this
				DerbyUserDAO.LOG.warn("Could not close query result.", ex);
			}
		}

		return result;
	}

}
